 
typedef struct{
    float2 start;
    float2 end;
} lineSegment;

typedef struct{
    float2 origin;
    float2 size;
    int count;
    int children;
    int lines[6];
} node;

////------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------

float3 src2rgb(float src) {
    
    float blueLimit = 0.6f;
    float redLimit = 0.0f;
    float range = blueLimit-redLimit;
    // map to the HSV range
    src *= range;
    src = range-src;
    src += redLimit;
    float hue = clamp(src,redLimit,blueLimit);
    
    float val = clamp((hue - 0.55f),0.0f, 0.1f);
    val *= 10;
    val = 1-val;
    
    float sat = 1;
    
    float r = 0;
    float g = 0;
    float b = 0;
    

    hue *= 360;
    // divide hue into six segments, 60 degree each
    int h_i = convert_int(floor(hue / 60)) % 6;
    float f = hue / 60 - floor(hue / 60);
    float p = val * (1 - sat);
    float q = val * (1 - f * sat);
    float t = val * (1 - (1 - f) * sat);

    switch (h_i) {
    case 0:
            r = val;
            g = t;
            b = p;
            break;
    case 1:
            r = q;
            g = val;
            b = p;
            break;
    case 2:
            r = p;
            g = val;
            b = t;
            break;
    case 3:
            r = p;
            g = q;
            b = val;
            break;
    case 4:
            r = t;
            g = p;
            b = val;
            break;
    case 5:
            r = val;
            g = p;
            b = q;
            break;
    }
    
    return  (float3)(r, g, b);;
 }

float2 LineIntersection(float2 ray_Start,  float2 ray_End, float2 line_Start, float2 line_End){
    float2 ray_Diff = ray_End-ray_Start;
    float  ray_Compare = (ray_Diff.x*ray_Start.y) - (ray_Diff.y*ray_Start.x);

    float2 intersection;
    intersection.x = -1;
    intersection.y = -1;
    float2 line_Diff = line_End - line_Start;
    float line_Compare = (line_Diff.x*line_Start.y) - (line_Diff.y*line_Start.x);
    
    if (((ray_Diff.x*line_Start.y - ray_Diff.y*line_Start.x ) < ray_Compare )
        ^ (( ray_Diff.x*line_End.y - ray_Diff.y*line_End.x ) < ray_Compare )){
        
        if(((line_Diff.x*ray_Start.y - line_Diff.y*ray_Start.x ) < line_Compare )
         ^ (( line_Diff.x*ray_End.y - line_Diff.y*ray_End.x) < line_Compare )){
            
            float lDetDivInv = 1 / ((ray_Diff.x*line_Diff.y) - (ray_Diff.y*line_Diff.x));
            intersection.x =  -((ray_Diff.x*line_Compare) - (ray_Compare*line_Diff.x)) * lDetDivInv ;
            intersection.y =  -((ray_Diff.y*line_Compare) - (ray_Compare*line_Diff.y)) * lDetDivInv ;
        }}
    return intersection;
}

float range_Finder(float2 intersection, float2 origin){
    float diffX = intersection.x - origin.x;
    float diffY = intersection.y - origin.y;
    float rangeOut = ((diffX * diffX) + (diffY * diffY));
    return rangeOut;
}

float2 searchOctree(__global node* octree, __global lineSegment* plan, __global int* planTypes, float2 ray_start, float2 ray_end, float2 planMode, int isoHorizon, int searchType){
    
    float2 result = (float2)(-1,-1);
    float2 windowResult = (float2)(-1,-1);
    float4 mirrorResult = (float4)(-1,-1,-1,-1);
    
    int width = octree[0].size.x;
    int height = octree[0].size.y;
    float2 origin = octree[0].origin;
    
    float2 ray_position = ray_start;
    float2 ray_norm = normalize(ray_end-ray_start);
    
    int a = 0;
    float closest = 0;
    bool loop_Octree = true;
    bool hasReflected = false;
    int escape = 0;
    int throughWindow = 0;
    
    while(loop_Octree){
        escape++;
        node thisNode = octree[a];
        // if live node is not a leaf, i.e. if it has children
        if(thisNode.count==-1){
            float Ox = thisNode.origin.x+(thisNode.size.x/2);
            float Oy = thisNode.origin.y+(thisNode.size.y/2);
            int cX = 0;
            int cY = 0;
            
            if(ray_position.x>Ox) cX = 1;
            else cX = 0;
            if(ray_position.y>=Oy) cY = 1;
            else cY = 0;
            
            int cT = (cY*2)+cX;
            int children = thisNode.children;
            a = children + cT;
            }
        
        // if live node is a leaf (then also the point should be in it)
        else{
            bool intersecting = false;
            loop_Octree = false;
            bool mirrorIsClosest = false;
            int Ncount = thisNode.count;
            float2 rayBase = ray_position-ray_norm;
            float Ox = thisNode.origin.x;
            float Oy = thisNode.origin.y;
            float Dx = Ox+thisNode.size.x;
            float Dy = Oy+thisNode.size.y;
            
            //check for intersections between ray and lines in node
            for(int i = 0; i<Ncount; i++){
                int id = thisNode.lines[i];
                float2 pEnd = plan[id].end;
                float2 pStart = plan[id].start;
                float2 intersection = LineIntersection(rayBase,  ray_end, pStart, pEnd);
                
                // before recording, check that intersection is in the node
                if(intersection.x>=Ox-1){
                    if(intersection.x<=Dx+1){
                        if(intersection.y>=Oy-1){
                            if(intersection.y<=Dy+1){
                                int lineType = planTypes[id];
                                //when we see through windows, result does not get set by one
                                bool WindowHit = false;
                                if(planMode.x==1){
                                    if(lineType==1) WindowHit = true;
                                }
                                if(WindowHit){
                                    windowResult = intersection;
                                    throughWindow = -1;
                                }
                                // when we see anything else
                                else{
                                    float range = range_Finder(intersection, rayBase);
                                    bool goIntersect = false;
                                    if(range<closest)  goIntersect = true;
                                    else if (closest==0)goIntersect = true;
                                    if(goIntersect){
                                        if(intersecting==false) intersecting = true;
                                        closest=range;
                                        // if we've hit a mirror when they are on
                                        bool mirrorHit = false;
                                        if(planMode.y==1){
                                            if(lineType==2){
                                                if(hasReflected==false) mirrorHit = true;
                                            }}
                                        if(mirrorHit){
                                            mirrorResult.x = intersection.x;
                                            mirrorResult.y = intersection.y;
                                            mirrorResult.z = pStart.x;
                                            mirrorResult.w = pStart.y;
                                            mirrorIsClosest = true;
                                        }
                                        else{
                                            result.x = intersection.x;
                                            result.y = intersection.y;
                                            if(mirrorIsClosest) mirrorIsClosest = false;
                                        }}}}}}}}
            
            // if intersection not found, step to the next octree cell and begin over
            if(!intersecting){
                bool inNode = false;
                bool resetSearch = false;
                float2 intersection = (float2)(-5,-5);
                
                // before stepping, check that the end of the ray is not in the node
                if(ray_end.x>=Ox){
                    if(ray_end.x<=Dx){
                        if(ray_end.y>=Oy){
                            if(ray_end.y<=Dy){
                                inNode = true;
                            }}}}
                
                // if the ray end is not in the node, step to the next octree cell and begin over
                if(inNode == false){
                    bool checkBoundaries = true;
                    // first check to see if the ray position is aligned with a boundary of the node
                    // if it is, then ray / boundary intersection could be parallel
                    if(ray_position.x==Ox){
                        // is the ray end also aligned?
                        if(ray_end.x==Ox){
                            // tell us not to check the node boundaries later on
                            checkBoundaries = false;
                            // need to know if it is up or down; i.e. is ray_norm positive or negative?
                            if(ray_norm.y>0) intersection.y = Dy+1;
                            else intersection.y = Oy-1;
                            intersection.x = ray_position.x;
                        }}
                    else if(ray_position.x==Dx){
                        if(ray_end.x==Dx){
                            checkBoundaries = false;
                            if(ray_norm.y>0) intersection.y = Dy+1;
                            else intersection.y = Oy-1;
                            intersection.x = ray_position.x;
                        }}
                    if(ray_position.y==Oy){
                        if(ray_end.y==Oy){
                            checkBoundaries = false;
                            if(ray_norm.x>0) intersection.x = Dx+1;
                            else intersection.x = Ox-1;
                            intersection.y = ray_position.y;
                        }}
                    else if(ray_position.y==Dy){
                        if(ray_end.y==Dy){
                            checkBoundaries = false;
                            if(ray_norm.x>0) intersection.x = Dx+1;
                            else intersection.x = Ox-1;
                            intersection.y = ray_position.y;
                        }}
                    
                    // check ray against all node boundaries
                    if(checkBoundaries){
                        int iMult = 0;
                        float2 lineA;
                        float2 lineB;
                        
                        if(ray_norm.x>0) iMult = 1;
                        lineA.x = Ox+(thisNode.size.x*iMult);
                        lineA.y = Oy;
                        lineB.x = Ox+(thisNode.size.x*iMult);
                        lineB.y = Oy+thisNode.size.y;
                        intersection = LineIntersection(ray_position,  ray_end, lineA, lineB);
                        //no intersection found
                        iMult = 0;
                        if(intersection.x==-1){
                            if(ray_norm.y>0) iMult = 1;
                            lineA.x = Ox;
                            lineA.y = Oy+(thisNode.size.y*iMult);
                            lineB.x = Ox+thisNode.size.x;
                            lineB.y = Oy+(thisNode.size.y*iMult);
                            intersection = LineIntersection(ray_position,  ray_end, lineA, lineB);
                        }}}
            
                // if still within octree bounds, allow to increment to next node
                if(intersection.x>origin.x){
                    if(intersection.x<(origin.x+width)){
                        if(intersection.y>origin.y){
                            if(intersection.y<(origin.y+height)){
                                // prime to reset
                                resetSearch = true;
                            }}}}
                
                if(resetSearch){
                    // push past the intersection
                    ray_position = intersection+(ray_norm);
                    //if safely out of current node
                    bool isOut = false;
                    
                    if(ray_position.x<Ox)           isOut = true;
                    else if (ray_position.x>=Dx)    isOut = true;
                    else if (ray_position.y<Oy)     isOut = true;
                    else if (ray_position.y>=Dy)    isOut = true;
                    if(isOut){
                        loop_Octree = true;
                        //restart octree search from top
                        a = 0;
                        closest = 0;
                    }}
                
                else{
                    // no wall intersection has been found
                    // either the ray end point is in the node, or we are at the edge of the screen
                    // if we are seeing through windows
                    // and a window intersection has been found
                    // and it is the first search, i.e. not the global depths search
                    bool winda = false;
                    if(planMode.x==1){
                        if(windowResult.x>=0){
                            if(searchType == 0) winda = true;
                        }}
                    if(winda){
                        // has hit the edge of screen after going through a window
                        // so save the result beyond the window
                        windowResult += ray_norm*isoHorizon;
                        result.s0 = windowResult.x;
                        result.s1 = windowResult.y;
                    }
                    // else no wall intersection has been found
                    else{
                        //override with view if edge of quadtree or screen has been found
                        if(intersection.x>-5) intersection += ray_norm*isoHorizon;
                        //record -6 or -7 (through a window) if end of ray has been reached
                        else if(inNode){
                            intersection.x = -6;
                            intersection.y = (-6+throughWindow);
                        }
                        //else records -5 scores here if failed result
                        result.s0 = intersection.x;
                        result.s1 = intersection.y;
                    }
                    //break the loop
                }}
        
            if(mirrorIsClosest){
               if(hasReflected==false){
                   // mirrorResult will be preserved and so saves the reflection line and point of intersection
                   // need to reflect ray normal and restart search
                   float2 L0;
                   L0.x = mirrorResult.x;
                   L0.y = mirrorResult.y;
                   float2 L1;
                   L1.x = mirrorResult.z;
                   L1.y = mirrorResult.w;
                
                   float2 toMirror = ray_position - L0;
                   float2 normalMirror = L0-L1;
                
                   //normalise mirror
                   //need to turn perpendicular
                   float length = (float)sqrt( normalMirror.x*normalMirror.x + normalMirror.y*normalMirror.y );
                   if( length > 0 ) {
                       float _x = normalMirror.x;
                       normalMirror.x = -(normalMirror.y/length);
                       normalMirror.y = _x/length;
                   }
                   //calculate dot product
                   float dotProd = dot(normalMirror,toMirror);
                   float2 reflection = L0+(toMirror-2*dotProd*normalMirror);
                
                   ray_norm = normalize(L0 - reflection);
                   ray_position = L0 + (ray_norm);
                   ray_end = ray_position + (ray_norm*isoHorizon);
                
                   // loop the search
                   intersecting = true;
                   loop_Octree = true;
                   //restart octree search from top
                   a = 0;
                   closest = 0;
                   //elsewhere we need to turn the ray result back after one bounce so set a boolean
                   hasReflected = true;
               }}}
        
        if(escape>1000) loop_Octree = false;
    }
    
    // if the ray has reflected
    if(hasReflected){
        // reflect the results back to allow a straight line distance etc check
        // real world result point
        float2 resultR;
        resultR.x = result.s0;
        resultR.y = result.s1;
        //mirror surface result point
        float2 L0;
        L0.x = mirrorResult.x;
        L0.y = mirrorResult.y;
        //line end for mirroring function
        float2 L1;
        L1.x = mirrorResult.z;
        L1.y = mirrorResult.w;
        
        float2 toMirror = resultR - L0;
        float2 normalMirror = L0-L1;
        
        //normalise mirror
        normalMirror = normalize(normalMirror);
        //don't need to turn perpendicular
        //calculate dot product
        float dotProd = dot(normalMirror,toMirror);
        float2 resultInMirror = L0+(toMirror-2*dotProd*normalMirror);
        
        result.s0 = resultInMirror.x;
        result.s1 = resultInMirror.y;
    }
    return result;
}

void calculateLocals(float4 *localData, float8 *shapeData, float count, float2 pointFinal, float2 pointCoords, float r_Range){
    
    float localS0 = localData->s0;
    float localS2 = localData->s2;
    float localS3 = localData->s3;
    
    int fullsweep = 360; //360 steps per cycle
    float counter = (float)count;
    float cycles = (counter/fullsweep);
    
    //find length of radial sq
    float closest = r_Range;

    //area accumulation
    localS0 *= count-1;
    localS0 += closest;
    localS0 /= count;
    
    //compactness from average chord length
    float2 lastPoint;
    lastPoint.x = shapeData->s2;
    lastPoint.y = shapeData->s3;
    float edge = sqrt(range_Finder(lastPoint, pointFinal));
    shapeData->s4 += edge;          //accumulate perimeter
    
    edge *= edge;
    shapeData->s5 += edge;     //total overall sequential variance between radials squared
    
    //centroid or drift
    float2 centroidCalc;
    centroidCalc.x = shapeData->s0;
    centroidCalc.y = shapeData->s1;
    centroidCalc *= count-1;
    centroidCalc += pointFinal;
    centroidCalc /= count;
    shapeData->s0 = centroidCalc.x;
    shapeData->s1 = centroidCalc.y;
    
    //average var between centroid and edge
    float centroidVar = sqrt(range_Finder(centroidCalc, pointFinal));
    float centroidAv = shapeData->s7;
    centroidAv *= count-1;
    centroidAv += centroidVar;
    centroidAv /= count;
    shapeData->s7 = centroidAv;
    
    //greatest vista
    if(r_Range>(localS3*localS3)) localS3 = sqrt(r_Range);
    
    localData->s0 = localS0;                                                        //area
    
    float totalP = shapeData->s4;
    float perimeter = totalP/cycles;                                                //perimeter at this point
    if(counter>180)  localData->s1 = (4*M_PI*localS0)/((perimeter*perimeter)+1);     //isoperimetric ratio (compactness)
        
    localS2 = (sqrt(range_Finder(centroidCalc, pointCoords)))/centroidAv;           //drift as ratio of distance between me and centroid to average radial of space from centroid
    localData->s2 = localS2;
        
    localData->s3 = localS3;                                                        //vista
        
    shapeData->s6 = ((shapeData->s5/(cycles)))/(localS0*50);                        //total overall sequential radial variance per round squared and normalised by area
    }

float calculateSpectrum( float myData, float2 scaleUse, int measure, float maximum, float minimum, int type){
 
    float src;
    
    float minVal = (scaleUse.x * maximum)+minimum;
    float maxVal = scaleUse.y * maximum;
    float rangeVal = maxVal-minVal;
    src = myData-minVal;
    
    if(type==0){
        if(src<0) src = 0;
    }
    
    src /= rangeVal;
    
    if(type==0){
        if(src>1) src = 1;
        if(measure>5) src = 1-src;
    }
    return src;
}

//------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------

__kernel void raySearch( __global float2* posBuffer,  __global lineSegment* plan,
                        __global float4* localData,  __global float8* shapeData, __global float2* depthData, __global float2* depthBuffer,
                        __global node* octree,  const int count,  const float2 target, const float2 ray_dir,
                        __global int* indexer, const float4 indexOrigins, __global int* planTypes, const float2 planMode, __global int* stepper, __global float2* floodSeeds,
                        const int ticker, const int rounds, __global float4* linkers, const int stepsType, const int isoHorizon, const int flush, const float2 countdown, __global int* floodLimits){
    
    int id = get_global_id(0);
    const int group_id = get_global_id(0) / get_local_size(0);
    
    __global float2 *pos = &posBuffer[id];
    float2 pointCoords = *pos;
    
    float4 myLocal = localData[id];
    float8 myShape = shapeData[id];
    float2 myMeans = depthData[id];
    float2 myDepth = depthBuffer[id];
    float2 ray_norm = ray_dir;
    int trigger = count-ticker;
    
    int ind_X = (int)(indexOrigins.x);
    int ind_Y = (int)(indexOrigins.y);
    int ind_Z = (int)(indexOrigins.z);
    int ind_W = (int)(indexOrigins.w);
    int indexWidth = ind_Z-ind_X ;
    
    int Scan_width = octree[0].size.x;
    int Scan_height = octree[0].size.y;
    float rangeSum = Scan_width*Scan_height*0.7;
    
    float rangeLength = Scan_width+Scan_height;
    
    float2 ray_end = pointCoords+(ray_norm*rangeLength);
    int horizon = isoHorizon;
    int horizonSq = horizon*horizon;
    int floodJump = 16;
    
    //search the octree for the nearest intercept on the ray
    float2 result = searchOctree(octree, plan, planTypes, pointCoords, ray_end, planMode, horizon, 0);
    
    //ensure a zero result for the closest calculation
    float2 pointFinal = pointCoords;
    float2 pointOnMirror = pointCoords;
    float2 pointRoundCorner = pointCoords;
    //replace zeroed result if a real result has been found (lines or screen edge)
    if(result.s0>=0){
        // calculate distance and data values
        pointFinal.x = result.s0;
        pointFinal.y = result.s1;
        float r_Range = range_Finder(pointFinal,pointCoords);
        if(r_Range>horizonSq){
            pointFinal = pointCoords+(ray_norm*horizon);
            r_Range = horizonSq;
        }
        calculateLocals(&myLocal, &myShape, count, pointFinal, pointCoords, r_Range);
        myShape.s2 = pointFinal.x;
        myShape.s3 = pointFinal.y;
        }
    
    //step depth algorithm
    //check to see if target is visible when count is 1 i.e. start of run
    if(trigger==1){
        floodLimits[id] = 0;
        // save local depths to global mean depths before the wipe
        if(rounds>0){
            if(myDepth.x<25){
                float mean_visible = myMeans.x;
                mean_visible *= (rounds-1);
                mean_visible += myDepth.x;
                mean_visible /= rounds;
                myMeans.x = mean_visible;
            }
            else{
                float mean_visible = myMeans.x;
                float add_Vis = mean_visible;
                mean_visible *= (rounds-1);
                mean_visible += add_Vis;
                mean_visible /= rounds;
                myMeans.x = mean_visible;
            }
            if(myDepth.y<500000){
                float mean_metric = myMeans.y;
                mean_metric *= (rounds-1);
                mean_metric += myDepth.y;
                mean_metric /= rounds;
                myMeans.y = mean_metric;
            }
            else{
                float mean_metric = myMeans.y;
                float add_Met = mean_metric;
                mean_metric *= (rounds-1);
                mean_metric += add_Met;
                mean_metric /= rounds;
                myMeans.y = mean_metric;
            }
            myDepth.x = 10000000;
            myDepth.y = 10000000;
        }
        
        float howfar = range_Finder(target, pointCoords);
        if(howfar<horizonSq){
            // if a direct line of sight between point and target, returns -5
            float2 iso_result = searchOctree(octree, plan, planTypes, pointCoords, target, planMode, horizon, 1);
            // set zero depth if so
            if(iso_result.s0==-6){
                int n = indexWidth*(target.y-ind_Y)+target.x-ind_X;
                int iD = indexer[n];
                floodSeeds[id].x = iD;
                floodSeeds[id].y = 1;
                if(id == iD) myDepth.x = 0;
                else myDepth.x = 1;
                myDepth.y = sqrt(howfar);
                stepper[0]++;
                }}}
    
    // after start of run, normal flush for all types of stepping
    else if(flush==0){
        // first for normal flooding
        // if live point
        float floodStep = floodSeeds[id].y;
        if(floodStep>0){
            //if allowed to cycle
            if(floodLimits[id]<150){
                //record live
                stepper[0]++;
                // load my seed and begin the calcs
                int seedId = floodSeeds[id].x;
                float2 seedTarget = posBuffer[seedId];
                float seedRange = depthBuffer[seedId].y;
                float seedDepth = depthBuffer[seedId].x;
            
                float checkDepthTally = 0;
                float depthMatcher = 0;
                float metriMatcher = 10000000;
            
                //look at eight neighbour locations around me, at floodStep distance away
                for(int k = -1; k<2; k++){
                    for(int l = -1; l<2; l++){
                        if(((k==0)&&(l==0))==false){
                            int nx = pointCoords.x + k*floodStep;
                            if(nx>ind_X){ if(nx<ind_Z){
                                int ny = pointCoords.y + l*floodStep;
                                if(ny>ind_Y){ if(ny<ind_W){
                                    int n = indexWidth*(ny-ind_Y)+nx-ind_X;
                                    int iD = indexer[n];
                                    // if neighbour is not a wall
                                    if(iD>=0){
                                        // grab neighbour's saved info
                                        float2 neighbour = (float2)(nx,ny);
                                        float depth_found = depthBuffer[iD].x;
                                        float range_found = depthBuffer[iD].y;
                                            
                                        float dSend = depth_found;
                                        float rSend = range_found;
                                        float sSend;
                                        bool sendData = false;
                                        bool send_D = false;
                                        bool send_R = false;
                                            
                                        // calculate the distance to seed and add distance saved at seed
                                        float pass_Range = range_Finder(neighbour,seedTarget);
                                        float n_Range;
                                        float n_Depth;
                                        if(pass_Range<horizonSq){
                                            n_Range = sqrt(pass_Range);
                                            n_Range += seedRange;
                                            // if that range beats neighbour's info, save data
                                            if((range_found-n_Range)>5){
                                                rSend = n_Range;
                                                sendData = true;
                                                send_R = true;
                                            }
                                            //find the depth at seed plus one step
                                            n_Depth = seedDepth+1;
                                            // if that depth beats neighbour's info, save data
                                            if((n_Depth<depth_found)){
                                                dSend = n_Depth;
                                                sendData = true;
                                                send_D = true;
                                            }
                                                
                                            if(sendData){
                                                // if my neighbour is visible from my seed
                                                float2 lineOfSight = searchOctree(octree, plan, planTypes, neighbour, seedTarget, planMode, horizon, 1);
                                                if(lineOfSight.s0==-6) sSend = seedId;
                                                else sendData = false;
                                            }}
                                            
                                        // if neighbour wasn't visible from my seed
                                        if(sendData == false){
                                            if(floodStep==1){
                                                dSend = depth_found;
                                                rSend = range_found;
                                                // calculate the distance to my neighbour and add my saved distance
                                                n_Range = sqrt((float)((k*k)+(l*l)));
                                                n_Range += myDepth.y;
                                                // if that range beats neighbour's info, save data
                                                if((range_found-n_Range)>5){
                                                    rSend = n_Range;
                                                    sendData = true;
                                                    send_R = true;
                                                }
                                                //use my depth plus one step
                                                n_Depth = myDepth.x + 1;
                                                // if that depth beats neighbour's info, save data
                                                if((n_Depth<depth_found)){
                                                    dSend = n_Depth;
                                                    sendData = true;
                                                    send_D = true;
                                                }
                                                if(sendData)sSend = id;
                                            }}
                                            
                                        if(sendData){
                                            // there has been a disturbance in the force
                                            floodSeeds[iD].x = sSend;
                                            floodSeeds[iD].y = floodJump;
                                            if(send_D) depthBuffer[iD].x = dSend;
                                            if(send_R) depthBuffer[iD].y = rSend;
                                            floodLimits[iD]+=1;
                                        }
                                
                                        // before turning off, check I am not a gap
                                        if(floodStep==1){
                                            if((myDepth.x-depth_found)>25){
                                                myDepth.x = depth_found;
                                                myDepth.y = range_found;
                                            }
                                            else if(myDepth.x>depth_found){
                                                //record cells around that are higher
                                                if(depthMatcher==0) depthMatcher = depth_found;
                                                if(range_found<metriMatcher) metriMatcher = range_found;
                                                if(depthMatcher==depth_found) checkDepthTally+=1;
                                            }}}}}}}}}}

                // in the first round
                if(floodStep==floodJump){
                    // check against linkers
                    for(int i = 0; i<10; i++){
                        if(linkers[i].x>=0){
                            float2 linkA = (float2)(linkers[i].x,linkers[i].y);
                            float2 linkB = (float2)(linkers[i].z,linkers[i].w);
                            if((linkA.x == pointCoords.x)&&(linkA.y == pointCoords.y)){
                                int n = ((linkB.y-ind_Y)*indexWidth)+linkB.x-ind_X;
                                int iD = indexer[n];
                                if(iD>=0){
                                    float depth_found = depthBuffer[iD].x;
                                    float range_found = depthBuffer[iD].y;
                                    if(depth_found>myDepth.x){
                                        depthBuffer[iD].x = myDepth.x;
                                        floodSeeds[iD].x = iD;
                                        floodSeeds[iD].y = floodJump;
                                    }
                                    if(range_found>myDepth.y){
                                        depthBuffer[iD].y = myDepth.y;
                                        floodSeeds[iD].x = iD;
                                        floodSeeds[iD].y = floodJump;
                                    }
                                }}
                            else if((linkB.x == pointCoords.x)&&(linkB.y == pointCoords.y)){
                                int n = ((linkA.y-ind_Y)*indexWidth)+linkA.x-ind_X;
                                int iD = indexer[n];
                                if(iD>=0){
                                    float depth_found = depthBuffer[iD].x;
                                    float range_found = depthBuffer[iD].y;
                                    if(depth_found>myDepth.x){
                                        depthBuffer[iD].x = myDepth.x;
                                        floodSeeds[iD].x = iD;
                                        floodSeeds[iD].y = floodJump;
                                    }
                                    if(range_found>myDepth.y){
                                        depthBuffer[iD].y = myDepth.y;
                                        floodSeeds[iD].x = iD;
                                        floodSeeds[iD].y = floodJump;
                                    }}}}}}

                floodStep /= 2;
                if(floodStep<1){
                    //if cycle since primed is completed, turn off
                    floodSeeds[id].y = 0;
                    //if six cells are higher, switch me
                    if(checkDepthTally>=6){
                        myDepth.x = depthMatcher;
                        myDepth.y = metriMatcher;
                    }}
                else floodSeeds[id].y = floodStep;
                }}}
    
    else if(stepsType==2){
        if(flush<2){
            stepper[0]++;
            floodLimits[id] = 0;
            floodSeeds[id].y = 1024;
        }
        else{
            int seedId = floodSeeds[id].x;
            //find live depths
            float seedDepth = depthBuffer[seedId].x;
            if(seedDepth==countdown.x){
                //find awake points
                float floodStep = floodSeeds[id].y;
                if(floodStep>0){
                    // find allowed to cycle
                    if(floodLimits[id]<150){
                        //increment live count
                        stepper[0]++;
                        //load seed and do calca
                        float2 seedTarget = posBuffer[seedId];
                        float seedRange = depthBuffer[seedId].y;

                        float checkDepthTally = 0;
                        float depthMatcher = 0;
                        float metriMatcher = 10000000;
                
                        //look at eight neighbour locations around me, at floodStep distance away
                        for(int k = -1; k<2; k++){
                            for(int l = -1; l<2; l++){
                                if(((k==0)&&(l==0))==false){
                                    int nx = pointCoords.x + k*floodStep;
                                    if(nx>ind_X){ if(nx<ind_Z){
                                        int ny = pointCoords.y + l*floodStep;
                                        if(ny>ind_Y){  if(ny<ind_W){
                                            int n = indexWidth*(ny-ind_Y)+nx-ind_X;
                                            int iD = indexer[n];
                                            // if neighbour is not a wall
                                            if(iD>=0){
                                                // grab neighbour's saved info
                                                float2 neighbour = (float2)(nx,ny);
                                                float depth_found = depthBuffer[iD].x;
                                                float range_found = depthBuffer[iD].y;
                                                
                                                float dSend = depth_found;
                                                float rSend = range_found;
                                                float sSend = -1;
                                                bool sendData = false;
                                                bool send_D = false;
                                                bool send_R = false;
                                            
                                                // calculate the distance to seed and add distance saved at seed
                                                float pass_Range = range_Finder(neighbour,seedTarget);
                                                float n_Range;
                                                float n_Depth;
                                                if(pass_Range<horizonSq){
                                                    n_Range = sqrt(pass_Range);
                                                    n_Range += seedRange;
                                                    // if that range beats neighbour's info, save data
                                                    if((range_found-n_Range)>5){
                                                        rSend = n_Range;
                                                        sendData = true;
                                                        send_R = true;
                                                    }
                                                    //find the depth at seed plus one step
                                                    n_Depth = seedDepth+1;
                                                    // if that depth beats neighbour's info, save data
                                                    if((n_Depth<depth_found)){
                                                        dSend = n_Depth;
                                                        sendData = true;
                                                        send_D = true;
                                                    }
                                                    
                                                    if(sendData){
                                                        // if my neighbour is visible from my seed
                                                        float2 lineOfSight = searchOctree(octree, plan, planTypes, neighbour, seedTarget, planMode, horizon, 1);
                                                        if(lineOfSight.s0==-6){
                                                            floodSeeds[iD].x = seedId;
                                                            floodSeeds[iD].y = 1024;
                                                            if(send_D) depthBuffer[iD].x = dSend;
                                                            if(send_R) depthBuffer[iD].y = rSend;
                                                            floodLimits[iD]+=1;
                                                        }}}
                                                
                                                // before turning off, check I am not a gap
                                                if(floodStep==1){
                                                    if((myDepth.x-depth_found)>25){
                                                        myDepth.x = depth_found;
                                                        myDepth.y = range_found;
                                                    }
                                                    else if(myDepth.x>depth_found){
                                                        //record cells around that are higher
                                                        if(depthMatcher==0) depthMatcher = depth_found;
                                                        if(range_found<metriMatcher) metriMatcher = range_found;
                                                        if(depthMatcher==depth_found) checkDepthTally+=1;
                                                    }}}}}}}}}}

                        floodStep /= 2;
                        if(floodStep<1){
                            //if cycle since primed is completed, turn off
                            floodSeeds[id].y = 0;
                            //if six cells are higher, switch me
                            if(checkDepthTally>=6){
                                myDepth.x = depthMatcher;
                                myDepth.y = metriMatcher;
                            }}
                        else floodSeeds[id].y = floodStep;
                    }}}}}
    
    localData[id] = myLocal;
    shapeData[id] = myShape;
    depthData[id] = myMeans;
    depthBuffer[id] = myDepth;
}

//------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------

__kernel void colourCalc(__global float2* posBuffer, __global float2* graphBuffer, __global float4* cosBuffer,
                         __global float4* localData, __global float8* shapeData, __global float2* depthData, __global float2* depthBuffer,
                         __global float2* limitDials, const int measure,
                         const int graphForm, const int xAxis, const int yAxis, const float2 graphDims, const float graphCalibrator, const float graphSpread,
                         __global float2* polys, const int polySize, __global float* selects,
                         __global float2* minMax, const int renderType){
    
    int id = get_global_id(0);
    
    __global float2 *pos = &posBuffer[id];
    __global float2 *graphPos = &graphBuffer[id];
    __global float4 *cos = &cosBuffer[id];
    
    float src;
    
    if(graphForm==0){
        float myData;
        float myMaximum = 1;
        float myMinimum = 0;
        bool goMeasure = false;
        
        if(measure == 1){       goMeasure = true; myData = localData[id].x;}
        else if (measure == 2){ goMeasure = true; myData = localData[id].y;}
        else if (measure == 3){ goMeasure = true; myData = localData[id].z;}
        else if (measure == 4){ goMeasure = true; myData = shapeData[id].s6;}
        else if (measure == 5){ goMeasure = true; myData = localData[id].w;}
        else if (measure == 6){ myData = depthBuffer[id].y;}
        else if (measure == 7){ myData = depthBuffer[id].x;}
        else if (measure == 8){ myData = depthData[id].y;}
        else if (measure == 9){ myData = depthData[id].x;}
        if(goMeasure){
            myMinimum = minMax[measure-1].x;
            myMaximum = minMax[measure-1].y;
        }
        else if(measure==6) myMaximum = minMax[6].y;
        else if(measure==8){
            myMinimum = minMax[6].x;
            myMaximum = minMax[6].y;
        }

        float2 scaleUse = limitDials[measure-1];
        src = calculateSpectrum(myData, scaleUse, measure, myMaximum, myMinimum, 0);
    }
    else{
        float xData;
        float xMaximum = 1;
        float xMinimum = 0;
        bool goMeasure = false;
        
        if(xAxis == 1){       goMeasure = true; xData = localData[id].x;}
        else if (xAxis == 2){ goMeasure = true; xData = localData[id].y;}
        else if (xAxis == 3){ goMeasure = true; xData = localData[id].z;}
        else if (xAxis == 4){ goMeasure = true; xData = shapeData[id].s6;}
        else if (xAxis == 5){ goMeasure = true; xData = localData[id].w;}
        else if (xAxis == 6){ xData = depthBuffer[id].y;}
        else if (xAxis == 7){ xData = depthBuffer[id].x;}
        else if (xAxis == 8){ xData = depthData[id].y;}
        else if (xAxis == 9){ xData = depthData[id].x;}
        if(goMeasure){
            xMinimum = minMax[xAxis-1].x;
            xMaximum = minMax[xAxis-1].y;
        }
        else if(xAxis==6) xMaximum = minMax[6].y;
        else if(xAxis==8){
            xMinimum = minMax[6].x;
            xMaximum = minMax[6].y;
        }
    
        float yData;
        float yMaximum = 1;
        float yMinimum = 0;
        goMeasure = false;
        
        if(yAxis == 1){       goMeasure = true; yData = localData[id].x;}
        else if (yAxis == 2){ goMeasure = true; yData = localData[id].y;}
        else if (yAxis == 3){ goMeasure = true; yData = localData[id].z;}
        else if (yAxis == 4){ goMeasure = true; yData = shapeData[id].s6;}
        else if (yAxis == 5){ goMeasure = true; yData = localData[id].w;}
        else if (yAxis == 6){ yData = depthBuffer[id].y;}
        else if (yAxis == 7){ yData = depthBuffer[id].x;}
        else if (yAxis == 8){ yData = depthData[id].y;}
        else if (yAxis == 9){ yData = depthData[id].x;}
        if(goMeasure){
            yMinimum = minMax[yAxis-1].x;
            yMaximum = minMax[yAxis-1].y;
        }
        else if(yAxis==6) yMaximum = minMax[6].y;
        else if(yAxis==8){
            yMinimum = minMax[6].x;
            yMaximum = minMax[6].y;
        }
        
        float2 scaleUseA = limitDials[xAxis-1];
        float2 scaleUseB = limitDials[yAxis-1];
        
        float srcA = calculateSpectrum(xData, scaleUseA, xAxis, xMaximum, xMinimum, 1);
        float srcB = calculateSpectrum(yData, scaleUseB, yAxis, yMaximum, yMinimum, 1);
        
        src = srcB/srcA*graphCalibrator;
        if(src > 1)  src = 1/src;
        
        float2 see = *pos;
        
        if(graphForm==2){
            graphPos->x = srcA*graphDims.x;
            graphPos->y = ((1-srcB)*graphDims.y);
            see = *graphPos;
        }
        
        //now test for lasoo selection
        double xinters = 0;
        int N = polySize;
        int counter = 0;
        
        float2 p1 = polys[0];
        for (int i=1;i<N;i++) {
            float2 p2 = polys[i % N];
            if (see.y > min(p1.y,p2.y)) {
                if (see.y <= max(p1.y,p2.y)) {
                    if (see.x <= max(p1.x,p2.x)) {
                        if (p1.y != p2.y) {
                            xinters = (see.y-p1.y)*(p2.x-p1.x)/(p2.y-p1.y)+p1.x;
                            if (p1.x == p2.x || see.x <= xinters)
                                counter++;
                        }}}}
            p1 = p2;
        }
        if (counter % 2 != 0) selects[id] = 1;
    }
    
    if(graphForm>0){
        src = 1-src;
        double info = (double)src;
        double power = (double)graphSpread;
        src = powr(info,power);
        src = 1-src;
    }
    
    if(selects[id]==1){
        cos->x = 1;
        cos->y = 0;
        cos->z = 1;
    }
    else{
        if(renderType==3){
            if(measure < 5) src *= 5;
            else src *= 40;
            int band = fmod(src,2);
            if(band==0){
                cos->x = 1;
                cos->y = 1;
                cos->z = 0;
            }
            else{
                cos->x = 0;
                cos->y = 0;
                cos->z = 1;
            }}
        else{
            if(renderType==2) src = 1-src;
            // convert to RGB
            if(src<0)src = 0;
            float3 fragRGB = src2rgb(src);
            cos->x = fragRGB.x;
            cos->y = fragRGB.y;
            cos->z = fragRGB.z;
        }}
}

__kernel void vectorCalc( __global float4* localData, __global float8* shapeData, __global float2* depthData, __global float2* depthBuffer,
                         const int measure, __global float2* vectorLines, __global int* indexer, const float4 indexOrigins, __global float2* minMax){
    
    int id = get_global_id(0);
    
    if(id%2!=0){
        __global float2 *pos = &vectorLines[id-1];
        float2 vectorLine = *pos;
        float2 averageVector = (float2)(0,0);
        float maximum;
        
        int m = measure;
        maximum = minMax[m-1].y;
        
        int ind_X = (int)(indexOrigins.x);
        int ind_Y = (int)(indexOrigins.y);
        int ind_Z = (int)(indexOrigins.z);
        int ind_W = (int)(indexOrigins.w);
        int indexWidth = ind_Z-ind_X;
        int indexHeight = ind_W-ind_Y;
        int me = ((vectorLine.y-ind_Y)*indexWidth)+vectorLine.x-ind_X;
        int miD = indexer[me];
        float myVal;

        if(m==1)        myVal = localData[miD].x;        // connectivity
        else if(m==2)   myVal = localData[miD].y;        // compactness
        else if(m==3)   myVal = localData[miD].z;        // drift
        else if(m==4)   myVal = shapeData[miD].s6;       // occlusivity
        else if(m==5)   myVal = localData[miD].w;        // vista
        else if(m==6)   myVal = depthBuffer[miD].y;      // metric depth
        else if(m==7)   myVal = depthBuffer[miD].x;      // step depth
        else if(m==8)   myVal = depthData[miD].y;        // metric integ
        else if(m==9)   myVal = depthData[miD].x;        // visual integ

        //look at eight neighbour locations around me
        for(int k = -1; k<2; k++){
            for(int l = -1; l<2; l++){
                if(((k==0)&&(l==0))==false){
                    int nx = vectorLine.x + k*10;
                    if((nx>ind_X)&&(nx<ind_Z)){
                        int ny = vectorLine.y + l*10;
                        if((ny>ind_Y)&&(ny<ind_W)){
                            int n =  ((ny-ind_Y)*indexWidth)+nx-ind_X;
                            int iD = indexer[n];
                            // if neighbour is not a wall
                            if(iD>=0){
                                float val = 0;
                                // load the value relevant for the measure on show
                                if(m==1)        val = localData[iD].x;      // connectivity
                                else if(m==2)   val = localData[iD].y;      // compactness
                                else if(m==3)   val = localData[iD].z;      // drift
                                else if(m==4)   val = shapeData[iD].s6;      // occlusivity
                                else if(m==5)   val = localData[iD].w;      // vista
                                else if(m==6)   val = depthBuffer[iD].y;      // metric depth
                                else if(m==7)   val = depthBuffer[iD].x;      // step depth
                                else if(m==8)   val = depthData[iD].y;      // metric integ
                                else if(m==9)   val = depthData[iD].x;      // visual integ

                                // find difference with myself i.e. displacement
                                val -= myVal;
                                // create displacement coordinate
                                float2 addVector;
                                addVector.x = k*val;
                                addVector.y = l*val;
                                // add coordinate to average
                                averageVector += addVector;
                            }}}}}}

        if(m==1)        myVal /= maximum;         // connectivity
        else if(m==2)   myVal /= maximum;         // compactness
        else if(m==3)   myVal /= (maximum/3);     // drift (nice to be hairy)
        else if(m==4)   myVal /= 1;                 // occlusivity is normalised but scales
        else if(m==5)   myVal /= maximum;         // vista
        else if(m==6)   myVal /= 1000;              // metric depth
        else if(m==7)   myVal /= 10;                // step depth
        else if(m==8)   myVal /= 1000;              // metric integ
        else if(m==9)   myVal /= 10;                // visual integ

        if(myVal >= 3)          myVal = 3;
        else if(myVal <= -3)    myVal = -3;
        //// add average to start of vectorLine
        averageVector = normalize(averageVector);
        averageVector *= (50*myVal);
        if(m==1) vectorLine += averageVector;
        else vectorLine -= averageVector;
        //// save result down as end of vectorline
        vectorLines[id] = vectorLine;
    }
}


